/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.Map.Entry;

import org.jbpm.pvm.Execution;
import org.jbpm.pvm.internal.model.ObservableElementImpl;
import org.jbpm.pvm.model.Node;
import org.ow2.orchestra.definition.element.Catch;
import org.ow2.orchestra.definition.element.CompensationHandler;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.definition.element.EventHandler;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.ScopeRuntime;

/**
 * the runtime behaviour of the scope activity.
 *
 * @author Charles Souillard, Tom Baeyens
 * @version $Id: Scope.java 1683 2008-11-19 08:25:42Z porcherg $
 */
public class Scope extends AbstractActivity {
  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  /**
   * List of variable name that must be initialized at the beginning of the scope.
   * (variable with an inline from-spec).
   */
  protected List<String> listOfVariableToBeInitialized;
  /** isolated attribute of this scope. */
  protected boolean isolated = false;
  /** exitOnStandardFault attribute of this scope. */
  protected boolean exitOnStandardFault;
  /** list of variables of this scope. */
  protected Map<String, Variable> variables;
  /** list of eventHandlers of this scope. */
  protected List<EventHandler> eventHandlers;
  /** list of faultHandlers of this scope. */
  protected CompensationHandler compensationHandler;
  /** list of partnerLinks of this scope. */
  protected Map<String, PartnerLink> partnerLinks;
  /** list of messageExchanges of this scope. */
  protected Set<String> messageExchanges;
  /** list of correlationSets of this scope. */
  protected Map<String, CorrelationSet> correlationSets;

  protected List<Catch> catchs;
  /*
    <xsd:element ref="compensationHandler" minOccurs="0"/>
    <xsd:element ref="terminationHandler" minOccurs="0"/>
   */

  protected Scope() {
    super();
  }
  public Scope(final String uuid) {
    super(uuid);
    this.setName("Scope " + uuid);
  }

  /**
   * business logic.
   * @param execution execution
   */
  @SuppressWarnings("unchecked")
  @Override
  public void executeActivity(final BpelExecution execution) {
    final ScopeRuntime sr = new ScopeRuntime(this, execution, UUID.randomUUID().toString());
    execution.pushScopeRuntime(sr);
    if (this.listOfVariableToBeInitialized != null) {
      for (final String variableName : this.listOfVariableToBeInitialized) {
        execution.getVariableByName(variableName).initializeXmlType(execution);
      }
    }

    // Add additional variables (used in ForEach)
    final Map<String, Object> additionalVariables = (Map<String, Object>) execution
        .getVariable(BpelExecution.ADDITIONNAL_SCOPE_VARIABLES);
    if (additionalVariables != null) {
      for (final Entry<String, Object> var : additionalVariables.entrySet()) {
        execution.getVariableByName(var.getKey()).setValue(var.getValue(), false);
      }
      execution.removeVariable(BpelExecution.ADDITIONNAL_SCOPE_VARIABLES);
    }

    // Add these compensations handlers to the parent scope
    execution.getScopeRuntime().addCompensationHandlerInCurrentScope(this.getCompensationHandler());

    // get the current sequence node
    final Node node = execution.getNode();
    final Node firstNode = node.getNodes().get(0);
    // execute scope in a child execution
    execution.createExecution().execute(firstNode);
  }

  /**
   * Signal method.
   * @param execution execution
   * @param signal signal
   */
  @Override
  public void signal(final BpelExecution execution, final String signal, final Map<String, Object> parameters) {
    if (signal == null) {
      // this is the child execution returning
      // end child execution
      execution.end(Execution.STATE_ENDED);

      final BpelExecution scopeExecution = execution.getParent();
      // remove finished execution from parent
      scopeExecution.removeExecution(execution);

      // signal must be non-null to make the distinction with child
      // executions
      // in the first line of this method
      scopeExecution.signal("finished");
    } else if (signal.equals("finished")) {
      executeSources(execution);
      if (execution.isActive()) {

        final ObservableElementImpl parent = execution.getNode().getParent();
        if (parent != null && parent instanceof Node) {

          List<CompensationHandler> chsChilds = null;
          // we're done and the sequence is completed
          final ScopeRuntime childScope = execution.popScopeRuntime();


          if (childScope != null) {
            chsChilds = childScope.getCompensationHandlers();
          }

          final ScopeRuntime parentScope = execution.getScopeRuntime();

          // Add these compensations handlers to the parent scope
          if ((parentScope != null) && (chsChilds != null)) {
            parentScope.addCompensationHandlersInCurrentScope(chsChilds);
          }
        }
      }
    }
  }

  /**
   * Searched for a correlation having the given name.
   * Search first in this scope and if nothing is found, search in father scopes...
   * @param correlationSetName correlationSetName
   * @return the found correlation
   */
  public CorrelationSet findCorrelationSet(final String correlationSetName) {
    if (this.correlationSets != null && this.correlationSets.containsKey(correlationSetName)) {
      return this.correlationSets.get(correlationSetName);
    } else if (this.enclosingScope != null) {
      return this.enclosingScope.findCorrelationSet(correlationSetName);
    }
    //TODO : throw an exception
    return null;
  }
  /**
   * Searched for a partnerLink having the given name.
   * Search first in this scope and if nothing is found, search in father scopes...
   * @param partnerLinkName partnerLinkName
   * @return the found partnerLink
   */
  public PartnerLink findPartnerLink(final String partnerLinkName) {
    if (this.partnerLinks != null && this.partnerLinks.containsKey(partnerLinkName)) {
      return this.partnerLinks.get(partnerLinkName);
    } else if (this.enclosingScope != null) {
      return this.enclosingScope.findPartnerLink(partnerLinkName);
    }
    //TODO : throw an exception
    return null;
  }
  /**
   * Searched for a variable having the given name.
   * Search first in this scope and if nothing is found, search in father scopes...
   * @param variableName variableName
   * @return the found variable
   */
  public Variable findVariable(final String variableName) {
    if (this.variables != null && this.variables.containsKey(variableName)) {
      return this.variables.get(variableName);
    } else if (this.enclosingScope != null) {
      return this.enclosingScope.findVariable(variableName);
    }
    //TODO : throw an exception
    return null;
  }
  /**
   * Add a variable in the list of variables that the scope needs to initialize.
   * @param variableName variableName
   */
  public void addVariableToBeInitialized(final String variableName) {
    if (this.listOfVariableToBeInitialized == null) {
      this.listOfVariableToBeInitialized = new ArrayList<String>();
    }
    this.listOfVariableToBeInitialized.add(variableName);
  }

  public Variable getVariable(final String variableName) {
    return this.variables.get(variableName);
  }
  public Map<String, PartnerLink> getPartnerLinks() {
    return this.partnerLinks;
  }
  public Map<String, Variable> getVariables() {
    return this.variables;
  }
  public List<EventHandler> getEventHandlers() {
    return this.eventHandlers;
  }
  public void setEventHandlers(final List<EventHandler> eventHandlers) {
    this.eventHandlers = eventHandlers;
  }
  public boolean isExitOnStandardFault() {
    return this.exitOnStandardFault;
  }
  public void setExitOnStandardFault(final boolean exitOnStandardFault) {
    this.exitOnStandardFault = exitOnStandardFault;
  }
  public boolean isIsolated() {
    return this.isolated;
  }
  public void setIsolated(final boolean isolated) {
    this.isolated = isolated;
  }
  public void setPartnerLinks(final Map<String, PartnerLink> partnerLinks) {
    this.partnerLinks = partnerLinks;
  }
  public void setVariables(final Map<String, Variable> variables) {
    this.variables = variables;
  }
  public Set<String> getMessageExchanges() {
    return this.messageExchanges;
  }
  public void setMessageExchanges(final Set<String> messageExchanges) {
    this.messageExchanges = messageExchanges;
  }
  public Map<String, CorrelationSet> getCorrelationSets() {
    return this.correlationSets;
  }
  public void setCorrelationSets(final Map<String, CorrelationSet> correlationSets) {
    this.correlationSets = correlationSets;
  }

  public CompensationHandler getCompensationHandler() {
    return this.compensationHandler;
  }
  public void setCompensationHandler(final CompensationHandler compensationHandler) {
    this.compensationHandler = compensationHandler;
  }



  @Override
  public String toString() {
    return Scope.class.getName()
        + "[listOfVariableToBeInitialized: " + this.listOfVariableToBeInitialized
        + ", isolated: " + this.isolated
        + ", exitOnStandardFault: " + this.exitOnStandardFault
        + ", variables: " + this.variables
        + ", eventHandlers: " + this.eventHandlers
        + ", catchs: " + this.catchs
        + ", compensationHandlers: " + this.compensationHandler
        + ", partnerLinks: " + this.partnerLinks
        + ", messageExchanges: " + this.messageExchanges
        + ", correlationSets: " + this.correlationSets
        + "]";
  }

  public List<Catch> getCatchs() {
    return this.catchs;
  }

  public void setCatchs(final List<Catch> catchs) {
    this.catchs = catchs;
  }

}
