/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.jmx;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import javax.xml.namespace.QName;

import org.apache.tools.ant.Project;
import org.apache.tools.ant.taskdefs.Delete;
import org.jbpm.pvm.env.Environment;
import org.jbpm.pvm.env.EnvironmentFactory;
import org.jbpm.pvm.internal.cmd.Command;
import org.jbpm.pvm.internal.cmd.CommandService;
import org.ow2.novabpm.util.Misc;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.deployment.Deployer;
import org.ow2.orchestra.env.EnvFactoryRepository;
import org.ow2.orchestra.facade.data.ProcessDefinitionData;
import org.ow2.orchestra.facade.exception.InstanceNotFoundException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.exception.ProcessNotFoundException;
import org.ow2.orchestra.facade.jmx.Deployment;
import org.ow2.orchestra.facade.jmx.InstanceInfo;
import org.ow2.orchestra.facade.jmx.RemoteDeployerMBean;
import org.ow2.orchestra.facade.runtime.ActivityFullInstance;
import org.ow2.orchestra.facade.runtime.ActivityInstance;
import org.ow2.orchestra.facade.runtime.ActivityState;
import org.ow2.orchestra.facade.runtime.ProcessFullInstance;
import org.ow2.orchestra.facade.runtime.ProcessInstance;
import org.ow2.orchestra.facade.runtime.ProcessState;
import org.ow2.orchestra.facade.runtime.impl.ActivityInstanceImpl;
import org.ow2.orchestra.facade.runtime.impl.ProcessInstanceImpl;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.parsing.BpelImportParser;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.MutexRepository;
import org.ow2.orchestra.services.Querier;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.util.EnvTool;

public class RemoteDeployerImpl implements RemoteDeployerMBean {

  public RemoteDeployerImpl() {
  }

  protected CommandService getCommandService() {
    EnvironmentFactory environmentFactory = EnvFactoryRepository.get();
    Misc.badStateIfNull(environmentFactory, "environmentFactory is null.");

    return environmentFactory.get(CommandService.class);
  }

  public ProcessDefinitionData deploy(final Deployment bpelDeployment) {

    return getCommandService().execute(new Command<ProcessDefinitionData>() {

      public ProcessDefinitionData execute(final Environment environment) throws Exception {
        Deployer deployer = new Deployer();

        BpelProcess bpelProcess = deployer.deploy(bpelDeployment, environment);
        if (bpelProcess == null) {
          throw new OrchestraRuntimeException("Process is null");
        }
        return bpelProcess.getProcessDefinitionData();
      }
    });

  }

  public boolean undeploy(final QName processQName) {
    return getCommandService().execute(new Command<Boolean>() {

      public Boolean execute(final Environment environment) throws Exception {
        Deployer deployer = new Deployer();
        deployer.undeploy(processQName, environment);
        return true;
      }
    });
  }

  public List<ProcessDefinitionData> listProcess() {
    return getCommandService().execute(new Command<List<ProcessDefinitionData>>() {

      public List<ProcessDefinitionData> execute(final Environment environment) throws Exception {
        List<ProcessDefinitionData> result = new ArrayList<ProcessDefinitionData>();
        Repository repository = environment.get(Repository.class);
        for (BpelProcess process : repository.getProcesses()) {
          result.add(process.getProcessDefinitionData());
        }
        return result;
      }
    });
  }

  public ProcessInfo getProcess(final QName processQName) {
    return getCommandService().execute(new Command<ProcessInfo>() {

      public ProcessInfo execute(final Environment environment) throws Exception {
        ProcessInfo recProcess = null;
        Repository repository = environment.get(Repository.class);
        // TODO: add others info for the process !
        if (repository.getProcess(processQName) != null) {
          recProcess = new ProcessInfo();
          recProcess.setQname(repository.getProcess(processQName).getQName());
        }
        return recProcess;
      }
    });
  }

  public String printInstances(final BpelExecution execution, final int depth) {
    String s = "-";
    for (int i = 0; i < depth; i++) {
      s += "--";
    }
    s += ">" + execution.getUUID() + " (" + execution.getState() + "), node= " + execution.getNode() + "\n";
    if (execution.getExecutions() != null) {
      for (org.jbpm.pvm.Execution child : execution.getExecutions()) {
        s += printInstances((BpelExecution) child, depth + 1);
      }
    }
    return s;
  }

  public InstanceInfo getInstanceRecord(final BpelExecution execution) {
    InstanceInfo instRec = new InstanceInfo();
    instRec.setInstanceUId(execution.getUUID());
    instRec.setNodeName(execution.getNode().getName());
    instRec.setState(execution.getState());
    return instRec;
  }

  public Map<QName, List<String>> listInstanceStatus() {
    return getCommandService().execute(new Command<Map<QName, List<String>>>() {

      public Map<QName, List<String>> execute(final Environment environment) throws Exception {
        Map<QName, List<String>> result = new HashMap<QName, List<String>>();
        Repository repository = environment.get(Repository.class);
        for (BpelProcess process : repository.getProcesses()) {
          result.put(process.getQName(), new ArrayList<String>());
          List<BpelExecution> instances = repository.getInstances(process.getQName());
          if (instances != null) {
            for (BpelExecution execution : instances) {
              result.get(process.getQName()).add(printInstances(execution, 0));
            }
          }
        }
        return result;
      }
    });
  }

  public Map<QName, List<InstanceInfo>> listInstance() {
    return getCommandService().execute(new Command<Map<QName, List<InstanceInfo>>>() {

      public Map<QName, List<InstanceInfo>> execute(final Environment environment) throws Exception {
        Map<QName, List<InstanceInfo>> result = new HashMap<QName, List<InstanceInfo>>();
        Repository repository = environment.get(Repository.class);
        for (BpelProcess process : repository.getProcesses()) {
          System.out.println("*** RemoteDeplyer: process name= " + process.getQName());
          result.put(process.getQName(), new ArrayList<InstanceInfo>());
          List<BpelExecution> instances = repository.getInstances(process.getQName());
          if (instances != null) {
            for (BpelExecution execution : instances) {
              System.out.println("*** RemoteDeplyer: instance uid = " + getInstanceRecord(execution).getInstanceUId());
              result.get(process.getQName()).add(getInstanceRecord(execution));
            }
          }
        }
        return result;
      }
    });
  }

  public ProcessDefinitionData deployBar(final byte[] barFile) {
    try {
      // Create temp dir
      new File(System.getProperty("java.io.tmpdir")).mkdirs();
      File tmpDir = File.createTempFile("orch", null, null);
      tmpDir.delete();
      if (!tmpDir.mkdirs()) {
        throw new IOException("Cannot create the temporary directory '" + tmpDir + "'.");
      }

      // Unzip file
      InputStream in = new ByteArrayInputStream(barFile);
      ZipInputStream zipInput = new ZipInputStream(in);
      ZipEntry zipFile;
      while ((zipFile = zipInput.getNextEntry()) != null) {
        // write the file to the disk
        // Create file
        OutputStream outputFile = new FileOutputStream(tmpDir.getAbsolutePath() + File.separator + zipFile.getName());
        try {
          // Copy file content
          int c;
          while ((c = zipInput.read()) != -1) {
            outputFile.write(c);
          }
        } finally {
          outputFile.close();
        }
      }
      zipInput.close();
      in.close();

      // Read properties from the unzipped directory:
      Properties props = new Properties();
      InputStream propsIs = new FileInputStream(tmpDir.getAbsolutePath() + File.separator + "bar.properties");
      try {
        props.load(propsIs);
      } finally {
        propsIs.close();
      }

      String bpelFileName = props.getProperty("bpel");
      URL bpelFile = new File(tmpDir.getAbsolutePath() + File.separator + bpelFileName).toURL();

      String wsdlsCmd = props.getProperty("wsdls");
      List<URL> wsdlUrls = new ArrayList<URL>();
      if (wsdlsCmd != null) {
        for (String wsdl : wsdlsCmd.split(",")) {
          wsdlUrls.add(new File(wsdl).toURL());
        }
      }
      Deployment deployment = BpelImportParser.parseBpelFromUrl(bpelFile, wsdlUrls);

      // Deploy the process
      ProcessDefinitionData processName = deploy(deployment);

      // Delete temp dir
      Delete d = new Delete();
      d.setTaskName("delete");
      d.setProject(new Project());
      d.setDir(tmpDir);
      d.execute();

      return processName;
    } catch (Exception e) {
      // TODO: Handle exceptions.
      throw new OrchestraRuntimeException("Unable to deploy bar", e);
    }
  }

  public void initialize() {
    getCommandService().execute(new Command<Object>() {

      public Object execute(final Environment environment) throws Exception {
        Publisher publisher = environment.get(Publisher.class);
        Repository repository = environment.get(Repository.class);

        for (BpelProcess process : repository.getProcesses()) {
          MutexRepository.createProcessLock(process.getQName());
          publisher.republishServices(process, environment);
          for (BpelExecution bpelExecution : repository.getInstances()) {
            MutexRepository.createInstanceLock(bpelExecution.getUUID());
          }
        }
        return null;
      }
    });
  }

  private void verifyInstance(final ProcessInstanceUUID processUUID) throws InstanceNotFoundException {
    InstanceNotFoundException instanceNotFoundException = getCommandService().execute(
      new Command<InstanceNotFoundException>() {

        public InstanceNotFoundException execute(final Environment environment) {
          Querier querier = EnvTool.getQuerier();
          if(querier.getProcessInstance(processUUID) == null) {
            return new InstanceNotFoundException(processUUID);
          }
          return null;
        }

      });

    if (instanceNotFoundException != null) {
      throw instanceNotFoundException;
    }
  }

  public Set<ActivityInstance> findActivityInstances(final ProcessInstanceUUID processUUID) throws InstanceNotFoundException {
    verifyInstance(processUUID);
    return getCommandService().execute(new Command<Set<ActivityInstance>>() {

      public Set<ActivityInstance> execute(final Environment environment) throws Exception {
        Set<ActivityInstance> result = new HashSet<ActivityInstance>();
        Querier querier = EnvTool.getQuerier();
        querier.getProcessInstance(processUUID);
        for (ActivityFullInstance activityFullInstance : querier.findActivityInstances(processUUID)) {
          result.add(new ActivityInstanceImpl(activityFullInstance));
        }
        return result;
      }
    });
  }

  public Set<ActivityInstance> findActivityInstances(final ProcessInstanceUUID processUUID, final ActivityState activityState)
    throws InstanceNotFoundException {
    verifyInstance(processUUID);
    return getCommandService().execute(new Command<Set<ActivityInstance>>() {

      public Set<ActivityInstance> execute(final Environment environment) throws Exception {
        Set<ActivityInstance> result = new HashSet<ActivityInstance>();
        Querier querier = EnvTool.getQuerier();
        for (ActivityFullInstance activityFullInstance : querier.findActivityInstances(processUUID, activityState)) {
          result.add(new ActivityInstanceImpl(activityFullInstance));
        }
        return result;
      }
    });
  }

  public Set<ProcessInstance> findProcessInstances() {
    return getCommandService().execute(new Command<Set<ProcessInstance>>() {

      public Set<ProcessInstance> execute(final Environment environment) throws Exception {
        Set<ProcessInstance> result = new HashSet<ProcessInstance>();
        Querier querier = EnvTool.getQuerier();
        for (ProcessFullInstance processFullInstance : querier.findProcessInstances()) {
          result.add(new ProcessInstanceImpl(processFullInstance));
        }
        return result;
      }
    });
  }

  public Set<ProcessInstance> findProcessInstances(final ProcessDefinitionUUID processDefinitionUUID)
    throws ProcessNotFoundException {
    return getCommandService().execute(new Command<Set<ProcessInstance>>() {

      public Set<ProcessInstance> execute(final Environment environment) throws Exception {
        Set<ProcessInstance> result = new HashSet<ProcessInstance>();
        Querier querier = EnvTool.getQuerier();
        for (ProcessFullInstance processFullInstance : querier.findProcessInstances(processDefinitionUUID)) {
          result.add(new ProcessInstanceImpl(processFullInstance));
        }
        return result;
      }
    });
  }

  public Set<ProcessInstance> findProcessInstances(final ProcessDefinitionUUID processDefinitionUUID,
    final ProcessState processState) throws ProcessNotFoundException {
    return getCommandService().execute(new Command<Set<ProcessInstance>>() {

      public Set<ProcessInstance> execute(final Environment environment) throws Exception {
        Set<ProcessInstance> result = new HashSet<ProcessInstance>();
        Querier querier = EnvTool.getQuerier();
        for (ProcessFullInstance processFullInstance : querier.findProcessInstances(processDefinitionUUID, processState)) {
          result.add(new ProcessInstanceImpl(processFullInstance));
        }
        return result;
      }
    });
  }

  public Set<ProcessInstance> findProcessInstances(final ProcessState processState) {
    return getCommandService().execute(new Command<Set<ProcessInstance>>() {

      public Set<ProcessInstance> execute(final Environment environment) throws Exception {
        Set<ProcessInstance> result = new HashSet<ProcessInstance>();
        Querier querier = EnvTool.getQuerier();
        for (ProcessFullInstance processFullInstance : querier.findProcessInstances(processState)) {
          result.add(new ProcessInstanceImpl(processFullInstance));
        }
        return result;
      }
    });
  }

}
